#===============================================================================
# Copyright 2014 NetApp, Inc. All Rights Reserved,
# contribution by Jorge Mora <mora@netapp.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#===============================================================================
"""
RPC Credentials module

Decode RPC Credentials.
"""
import gss
import gss_const
import rpc_const
import nfstest_config as c
from packet.utils import *
from baseobj import BaseObj

# Module constants
__author__    = "Jorge Mora (%s)" % c.NFSTEST_AUTHOR_EMAIL
__copyright__ = "Copyright (C) 2014 NetApp, Inc."
__license__   = "GPL v2"
__version__   = "1.2"

class auth_flavor(Enum):
    """enum auth_flavor"""
    _enumdict = rpc_const.auth_flavor

class rpc_gss_proc(Enum):
    """enum rpc_gss_proc"""
    _enumdict = gss_const.rpc_gss_proc

class rpc_gss_service(Enum):
    """enum rpc_gss_service"""
    _enumdict = gss_const.rpc_gss_service

class AuthNone(BaseObj):
    """AuthNone object"""
    # Class attributes
    flavor = auth_flavor(rpc_const.AUTH_NONE)
    _strfmt2  = "{0}"
    _attrlist = ("flavor",)

    def __init__(self, unpack):
        """Constructor which takes the Unpack object as input"""
        # Discard the length of data which should be 0
        unpack.unpack_uint()

class AuthSys(BaseObj):
    """AuthSys object"""
    # Class attributes
    flavor = auth_flavor(rpc_const.AUTH_SYS)
    _strfmt2  = "{0}({4}:{5})"
    _attrlist = ("flavor", "size", "stamp", "machine", "uid", "gid", "gids")

    def __init__(self, unpack):
        """Constructor which takes the Unpack object as input"""
        self.size    = unpack.unpack_uint()
        self.stamp   = unpack.unpack_uint()
        self.machine = unpack.unpack_opaque(maxcount=255)
        self.uid     = unpack.unpack_uint()
        self.gid     = unpack.unpack_uint()
        self.gids    = unpack.unpack_array(maxcount=16)

class GSS_Credential(BaseObj):
    """GSS_Credential object"""
    # Class attributes
    flavor = auth_flavor(rpc_const.RPCSEC_GSS)
    _strfmt2  = "{0}({3}:{5:@12})"
    _attrlist = ("flavor", "size", "gss_version", "gss_proc", "gss_seq_num",
                 "gss_service", "gss_context")

    def __init__(self, unpack):
        """Constructor which takes the Unpack object as input"""
        self.size        = unpack.unpack_uint()
        self.gss_version = unpack.unpack_uint()
        self.gss_proc    = rpc_gss_proc(unpack.unpack_uint())
        self.gss_seq_num = unpack.unpack_uint()
        self.gss_service = rpc_gss_service(unpack.unpack_uint())
        self.gss_context = unpack.unpack_opaque()

class GSS_Verifier(BaseObj):
    """GSS_Verifier object"""
    # Class attributes
    flavor = auth_flavor(rpc_const.RPCSEC_GSS)
    _strfmt2  = "{0}"
    _attrlist = ("flavor", "size", "gss_token")

    def __init__(self, unpack):
        """Constructor which takes the Unpack object as input"""
        self.gss_token = unpack.unpack_opaque()
        self.size      = len(self.gss_token)
        try:
            krb5 = gss.GSS_API(self.gss_token)
            if krb5:
                self.gss_token = krb5
        except:
            pass

def rpc_credential(unpack, verifier=False):
    """Process and return the credential or verifier"""
    try:
        # Get credential/verifier flavor
        flavor = unpack.unpack_uint()
        if flavor == rpc_const.AUTH_SYS:
            return AuthSys(unpack)
        elif flavor == rpc_const.AUTH_NONE:
            return AuthNone(unpack)
        elif flavor == rpc_const.RPCSEC_GSS:
            if verifier:
                return GSS_Verifier(unpack)
            else:
                return GSS_Credential(unpack)
    except:
        return None



#===============================================================================
# Copyright 2012 NetApp, Inc. All Rights Reserved,
# contribution by Jorge Mora <mora@netapp.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#===============================================================================
"""
TCP module

Decode TCP layer.

RFC  793 TRANSMISSION CONTROL PROTOCOL
RFC 2018 TCP Selective Acknowledgment Options
RFC 7323 TCP Extensions for High Performance
"""
import nfstest_config as c
from baseobj import BaseObj
from packet.unpack import Unpack
from packet.application.dns import DNS
from packet.application.rpc import RPC
from packet.application.krb5 import KRB5
from packet.utils import OptionFlags, ShortHex

# Module constants
__author__    = "Jorge Mora (%s)" % c.NFSTEST_AUTHOR_EMAIL
__copyright__ = "Copyright (C) 2012 NetApp, Inc."
__license__   = "GPL v2"
__version__   = "1.5"

TCPflags = {
    0: "FIN",
    1: "SYN",
    2: "RST",
    3: "PSH",
    4: "ACK",
    5: "URG",
    6: "ECE",
    7: "CWR",
    8: "NS",
}

class Stream(BaseObj):
    """TCP stream buffer object"""
    # Printing of this object is used for debugging only so don't display buffer
    _attrlist = ("last_seq", "next_seq", "seq_wrap", "seq_base", "frag_off")

    def __init__(self, seqno):
        self.buffer   = "" # Keep track of RPC packets spanning multiple TCP packets
        self.frag_off = 0  # Keep track of multiple RPC packets within a single TCP packet
        self.last_seq = 0  # Last sequence number processed
        self.next_seq = 0  # Next sequence number expected
        self.seq_wrap = 0  # Keep track when sequence number has wrapped around
        self.seq_base = seqno # Base sequence number to convert to relative sequence numbers
        self.segments = [] # Array of missing fragments, item: [start seq, end seq]

    def add_fragment(self, data, seq):
        """Add fragment data to stream buffer"""
        if len(data) == 0:
            return
        if seq == self.next_seq or len(self.buffer) == 0:
            # Append fragment data to stream buffer
            self.buffer += data
            self.segments = []
        elif seq > self.next_seq:
            # Previous fragment is missing so fill previous fragment with zeros
            size = seq - self.next_seq
            self.segments.append([self.next_seq, seq])
            self.buffer += "\x00" * size
            self.buffer += data
        else:
            # Fragment is out of order -- found previous missing fragment
            off  = len(self.buffer) - self.next_seq + seq
            datalen = len(data)
            size = datalen + off
            # Insert fragment where it belongs
            self.buffer = self.buffer[:off] + data + self.buffer[size:]

            # Remove fragment from segments list
            index = 0
            for frag in self.segments:
                if seq >= frag[0] and seq < frag[1]:
                    if seq == frag[0] and seq+datalen == frag[1]:
                        # Full segment matched, so just remove it
                        self.segments.pop(index)
                    elif seq == frag[0]:
                        # Start of segment matched, set new missing start
                        frag[0] = seq+datalen
                    elif seq+datalen == frag[1]:
                        # End of segment matched, set new missing end
                        frag[1] = seq
                    else:
                        # Full segment is within missing segment,
                        # set new missing end and create a new segment
                        newfrag = [seq+datalen, frag[1]]
                        frag[1] = seq
                        self.segments.insert(index+1, newfrag)
                    break
                index += 1

    def missing_fragment(self, seq):
        """Check if given sequence number is within a missing fragment"""
        for frag in self.segments:
            if seq >= frag[0] and seq < frag[1]:
                return True
        return False

class Flags(OptionFlags):
    """TCP Option flags"""
    _rawfunc  = ShortHex
    _bitnames = TCPflags
    __str__ = OptionFlags.str_flags

class Option(BaseObj):
    """Option object"""
    def __init__(self, unpack):
        """Constructor which takes an unpack object as input"""
        self.kind = None
        try:
            self.kind = unpack.unpack_uchar()
            if self.kind not in (0,1):
                length = unpack.unpack_uchar()
                if length > 2:
                    if self.kind == 2:
                        # Maximum Segment Size (MSS)
                        self.mss = unpack.unpack_ushort()
                        self._attrlist = ("kind", "mss")
                    elif self.kind == 3:
                        # Window Scale option (WSopt)
                        self.wsopt = unpack.unpack_uchar()
                        self._attrlist = ("kind", "wsopt")
                    elif self.kind == 5:
                        # Sack Option Format
                        self.blocks = []
                        for i in range((length-2)/8):
                            left_edge  = unpack.unpack_uint()
                            right_edge = unpack.unpack_uint()
                            self.blocks.append([left_edge, right_edge])
                        self._attrlist = ("kind", "blocks")
                    elif self.kind == 8:
                        # Timestamps option (TSopt)
                        self.tsval = unpack.unpack_uint()
                        self.tsecr = unpack.unpack_uint()
                        self._attrlist = ("kind", "tsval", "tsecr")
                    else:
                        self.data = unpack.read(length-2)
                        self._attrlist = ("kind", "data")
        except:
            pass

class TCP(BaseObj):
    """TCP object

       Usage:
           from packet.transport.tcp import TCP

           x = TCP(pktt)

       Object definition:

       TCP(
           src_port    = int, # Source port
           dst_port    = int, # Destination port
           seq_number  = int, # Sequence number
           ack_number  = int, # Acknowledgment number
           hl          = int, # Data offset or header length (32bit words)
           header_size = int, # Data offset or header length in bytes
           flags = Flags(     # TCP flags:
               rawflags = int,#   Raw flags
               FIN = int,     #   No more data from sender
               SYN = int,     #   Synchronize sequence numbers
               RST = int,     #   Synchronize sequence numbers
               PSH = int,     #   Push function. Asks to push the buffered
                              #     data to the receiving application
               ACK = int,     #   Acknowledgment field is significant
               URG = int,     #   Urgent pointer field is significant
               ECE = int,     #   ECN-Echo has a dual role:
                              #     SYN=1, the TCP peer is ECN capable.
                              #     SYN=0, packet with Congestion Experienced
                              #     flag in IP header set is received during
                              #     normal transmission
               CWR = int,     #   Congestion Window Reduced
               NS  = int,     #   ECN-nonce concealment protection
           ),
           window_size = int, # Window size
           checksum    = int, # Checksum
           urgent_ptr  = int, # Urgent pointer
           seq         = int, # Relative sequence number
           options = list,    # List of TCP options
           data = string,     # Raw data of payload if unable to decode
       )
    """
    # Class attributes
    _attrlist = ("src_port", "dst_port", "seq_number", "ack_number", "hl",
                 "header_size", "flags", "window_size", "checksum",
                 "urgent_ptr", "options", "length", "data")

    def __init__(self, pktt):
        """Constructor

           Initialize object's private data.

           pktt:
               Packet trace object (packet.pktt.Pktt) so this layer has
               access to the parent layers.
        """
        # Decode the TCP layer header
        unpack = pktt.unpack
        ulist = unpack.unpack(20, "!HHIIHHHH")
        self.src_port    = ulist[0]
        self.dst_port    = ulist[1]
        self.seq_number  = ulist[2]
        self.ack_number  = ulist[3]
        self.hl          = ulist[4] >> 12
        self.header_size = 4*self.hl
        self.flags       = Flags(ulist[4] & 0x1FF)
        self.window_size = ulist[5]
        self.checksum    = ShortHex(ulist[6])
        self.urgent_ptr  = ulist[7]

        pktt.pkt.tcp = self

        # Stream identifier
        ip = pktt.pkt.ip
        streamid = "%s:%d-%s:%d" % (ip.src, self.src_port, ip.dst, self.dst_port)

        if streamid not in pktt._tcp_stream_map:
            # Create TCP stream object
            pktt._tcp_stream_map[streamid] = Stream(self.seq_number)

        # De-reference stream map
        stream = pktt._tcp_stream_map[streamid]

        if self.flags.SYN:
            # Reset seq_base on SYN
            stream.seq_base = self.seq_number
            stream.last_seq = stream.seq_wrap

        # Convert sequence numbers to relative numbers
        seq = self.seq_number - stream.seq_base + stream.seq_wrap
        if seq < stream.seq_wrap:
            # Sequence number has reached the maximum and wrapped around
            stream.seq_wrap += 4294967296
            seq += 4294967296
        self.seq = seq

        if self.header_size > 20:
            self.options = []
            osize = self.header_size - 20
            optunpack = Unpack(unpack.read(osize))
            while optunpack.size():
                optobj = Option(optunpack)
                if optobj.kind == 0:
                    # End of option list
                    break
                elif optobj.kind > 0:
                    # Valid option
                    self.options.append(optobj)

        # Save length of TCP segment
        self.length = unpack.size()

        if seq < stream.last_seq and not stream.missing_fragment(seq):
            # This is a re-transmission, do not process
            return

        self._decode_payload(pktt, stream)

        if self.length > 0:
            stream.last_seq = seq
            stream.next_seq = seq + self.length

    def __str__(self):
        """String representation of object

           The representation depends on the verbose level set by debug_repr().
           If set to 0 the generic object representation is returned.
           If set to 1 the representation of the object is condensed:
               'TCP 708 -> 2049, seq: 3294175829, ack: 3395739041, ACK,FIN'

           If set to 2 the representation of the object also includes the
           length of payload and a little bit more verbose:
               'src port 708 -> dst port 2049, seq: 3294175829, ack: 3395739041, len: 0, flags: FIN,ACK'
        """
        rdebug = self.debug_repr()
        if rdebug == 1:
            out = "TCP %d -> %d, seq: %d, ack: %d, %s" % \
                  (self.src_port, self.dst_port, self.seq_number, self.ack_number, self.flags)
        elif rdebug == 2:
            out = "src port %d -> dst port %d, seq: %d, ack: %d, len: %d, flags: %s" % \
                  (self.src_port, self.dst_port, self.seq_number, self.ack_number, self.length, self.flags)
        else:
            out = BaseObj.__str__(self)
        return out

    def _decode_payload(self, pktt, stream):
        """Decode TCP payload."""
        rpc = None
        pkt = pktt.pkt
        unpack = pktt.unpack

        if 53 in [self.src_port, self.dst_port]:
            # DNS on port 53
            dns = DNS(pktt, proto=6)
            if dns:
                pkt.dns = dns
            return
        elif 88 in [self.src_port, self.dst_port]:
            # KRB5 on port 88
            krb = KRB5(pktt, proto=6)
            if krb:
                pkt.krb = krb
            return

        if stream.frag_off > 0 and len(stream.buffer) == 0:
            # This RPC packet lies within previous TCP packet,
            # Re-position the offset of the data
            unpack.seek(unpack.tell() + stream.frag_off)

        # Get the total size
        sid = unpack.save_state()
        size = unpack.size()

        # Try decoding the RPC header before using the stream buffer data
        # to re-sync the stream
        if len(stream.buffer) > 0:
            rpc = RPC(pktt, proto=6)
            if not rpc:
                unpack.restore_state(sid)
                sid = unpack.save_state()

        if rpc or (size == 0 and len(stream.buffer) > 0 and self.flags.rawflags != 0x10):
            # There has been some data lost in the capture,
            # to continue decoding next packets, reset stream
            # except if this packet is just a TCP ACK (flags = 0x10)
            stream.buffer = ""
            stream.frag_off = 0

        if not rpc:
            if len(stream.buffer):
                # Concatenate previous fragment
                unpack.insert(stream.buffer)
            ldata = unpack.size() - 4

            # Get RPC header
            rpc = RPC(pktt, proto=6)
        else:
            ldata = size - 4

        if not rpc:
            return

        rpcsize = rpc.fragment_hdr.size

        truncbytes = pkt.record.length_orig - pkt.record.length_inc
        if truncbytes == 0 and ldata < rpcsize:
            # An RPC fragment is missing to decode RPC payload
            unpack.restore_state(sid)
            stream.add_fragment(unpack.getbytes(), self.seq)
        else:
            if len(stream.buffer) > 0 or ldata == rpcsize:
                stream.frag_off = 0
            stream.buffer = ""
            # Save RPC layer on packet object
            pkt.rpc = rpc
            if rpc.type:
                # Remove packet call from the xid map since reply has
                # already been decoded
                pktt._rpc_xid_map.pop(rpc.xid, None)

            # Decode NFS layer
            rpcload = rpc.decode_payload()
            rpcbytes = ldata - unpack.size()
            if not rpcload and rpcbytes != rpcsize:
                pass
            elif unpack.size():
                # Save the offset of next RPC packet within this TCP packet
                # Data offset is cumulative
                stream.frag_off += size - unpack.size()
                sid = unpack.save_state()
                ldata = unpack.size() - 4
                try:
                    rpc_header = RPC(pktt, proto=6, state=False)
                except Exception:
                    rpc_header = None
                if not rpc_header or ldata < rpc_header.fragment_hdr.size:
                    # Part of next RPC packet is within this TCP packet
                    # Save the multi-span fragment data
                    unpack.restore_state(sid)
                    stream.add_fragment(unpack.getbytes(), self.seq)
                else:
                    # Next RPC packet is entirely within this TCP packet
                    # Re-position the file pointer to the current offset
                    pktt.seek(pktt.boffset)
            else:
                stream.frag_off = 0



#===============================================================================
# Copyright 2012 NetApp, Inc. All Rights Reserved,
# contribution by Jorge Mora <mora@netapp.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#===============================================================================
"""
Base object

Base class so objects will inherit the methods providing the string
representation of the object and methods to change the verbosity of such
string representation. It also includes a simple debug printing and logging
mechanism including methods to change the debug verbosity level and methods
to add debug levels.
"""
import re
import time
import nfstest_config as c
from pprint import pformat
from formatstr import FormatStr

# Module constants
__author__    = "Jorge Mora (%s)" % c.NFSTEST_AUTHOR_EMAIL
__copyright__ = "Copyright (C) 2012 NetApp, Inc."
__license__   = "GPL v2"
__version__   = "1.2"

# Module variables
_dindent = ""
_sindent = "    "
_dlevel = 0
_rlevel = 1
_dcount = 0
_strsize = 0
_logfh = None
_tstamp = True
_tstampfmt = "{0:date:%H:%M:%S.%q - }"

# Simple verbose level names
_debug_map = {
    'none':  0,
    'info':  1,    # Display info only
    'debug': 0xFF, # Display info and all debug messages 0x02-0x80
    'all':   0xFFFFFFFF, # Display all messages
}
# Debug display prefixes
_debug_prefix = {
    0x001: 'INFO: ',
}

def _init_debug():
    """Define all debug flags"""
    for i in xrange(7):
        dbg = 'dbg%d' % (i+1)
        _debug_map[dbg] = (2 << i)
        _debug_prefix[(2 << i)] = dbg.upper() + ': '
_init_debug()

# Instantiate FormatStr object
fstrobj = FormatStr()

class BaseObj(object):
    """Base class so objects will inherit the methods providing the string
       representation of the object and a simple debug printing and logging
       mechanism.

       Usage:
           from baseobj import BaseObj

           # Named arguments
           x = BaseObj(a=1, b=2)

           # Dictionary argument
           x = BaseObj({'a':1, 'b':2})

           # Tuple arguments: first for keys and second for the values
           x = BaseObj(['a', 'b'], [1, 2])

           # All of the above will create an object having two attributes:
           x.a = 1 and x.b = 2

           # Add attribute name, this will be the only attribute to be displayed
           x.set_attrlist("a")

           # Add list of attribute names to be displayed in that order
           x.set_attrlist(["a", "b"])

           # Set attribute with ordered display rights
           x.set_attr("a", 1)
           # This is the same as
           setattr(x, "a", 1) or x.a = 1
           x.set_attrlist("a")

           # Set attribute with switch duplicate
           # The following creates an extra attribute "switch" with
           # the same value as attribute "a":
           #   x.a == x.switch
           #   x.a is x.switch
           x.set_attr("a", 1, switch=True)

           # Make the current object flat by allowing all the attributes
           # for the new attribute to be accessed directly by the current
           # object so the following is True:
           #   x.d == x.c.d
           x.set_attr("c", BaseObj(d=11, e=22), switch=True)

           # Set the comparison attribute so x == x.a is True
           x.set_eqattr("a")

           # Set verbose level of object's string representation
           x.debug_repr(level)

           # Set string format for verbose level 1
           x.set_strfmt(1, "arg1:{0}")
           # In the above example the first positional argument is "a"
           # so the str(x) gives "arg1:1"

           # Set attribute shared by all instances
           # If a global or shared attribute is set on one instance,
           # all other instances will have access to it:
           #   y = BaseObj(d=2, e=3)
           # then the following is true
           #   x.g == y.g
           #   x.g is y.g
           x.set_global("g", 5)

           # Set level mask to display all debug messages matching mask
           x.debug_level(0xFF)

           # Add a debug mapping for mask 0x100
           x.debug_map(0x100, 'opts', "OPTS: ")

           # Set global indentation to 4 spaces for dprint
           x.dindent(4)

           # Set global indentation to 4 spaces for displaying objects
           x.sindent(4)

           # Set global truncation to 64 for displaying string objects
           x.strsize(64)

           # Do not display timestamp for dprint messages
           x.tstamp(enable=False)

           # Change timestamp format to include the date
           x.tstamp(fmt="{0:date:%Y-%m-%d %H:%M:%S.%q} ")

           # Get timestamp if enabled, else return an empty string
           out = x.timestamp()

           # Open log file
           x.open_log(logfile)

           # Close log file
           x.close_log()

           # Write data to log file
           x.write_log(data)

           # Format the given arguments
           out = x.format("{0:x} - {1}", 1, "hello")

           # Format the object attributes set by set_attrlist()
           out = x.format("{0:x} - {1}")

           # Print debug message only if OPTS bitmap matches the current
           # debug level mask
           x.dprint("OPTS", "This is an OPTS debug message")
    """
    # Class attributes
    _attrlist = None # List of attributes to display in order
    _eqattr   = None # Comparison attribute
    _attrs    = None # Dictionary where the key becomes an attribute which is
                     # a reference to another attribute given by its value
    _fattrs   = None # Make the object attributes of each of the attributes
                     # listed part of the attributes of the current object
    _strfmt1  = None # String format for verbose level 1
    _strfmt2  = None # String format for verbose level 2
    _globals  = {}   # Attributes share by all instances

    def __init__(self, *kwts, **kwds):
        """Constructor

           Initialize object's private data according to the arguments given.
           Arguments can be given as positional, named arguments or a
           combination of both.
        """
        keys = None
        for item in kwts:
            if type(item) == dict:
                self.__dict__.update(item)
            elif type(item) == list or type(item) == tuple:
                if keys is None:
                    keys = item
                else:
                    self.__dict__.update(zip(keys,item))
                    keys = None
        # Process named arguments: x = BaseObj(a=1, b=2)
        self.__dict__.update(kwds)

    def __getattr__(self, attr):
        """Return the attribute value for which the lookup has not found
           the attribute in the usual places. It checks the internal
           dictionary for any attribute references, it checks if this
           is a flat object and returns the appropriate attribute.
           And finally, if any of the attributes listed in _attrlist
           does not exist it returns None as if they exist but not
           defined
        """
        if self._globals.has_key(attr):
            # Shared attribute
            return self._globals[attr]
        if self._attrs is not None:
            # Check if attribute is a reference to another attribute
            name = self._attrs.get(attr)
            if name is not None:
                return getattr(self, name)
        if self._fattrs is not None:
            # Check if this is defined as a flat object so any attributes
            # of sub-objects pointed to by _fattrs are treated like
            # attributes of this object
            for item in self._fattrs:
                obj = getattr(self, item, None)
                if obj is not None and hasattr(obj, attr):
                    # Flat object: sub-object attributes as object attribute
                    return getattr(obj, attr)
        if self._attrlist is not None and attr in self._attrlist:
            # Make all attributes listed in _attrlist available even if they
            # haven't been defined
            return None
        raise AttributeError("'%s' object has no attribute '%s'" % (self.__class__.__name__, attr))

    def __eq__(self, other):
        """Comparison method: this object is treated like the attribute
           defined by set_eqattr()
        """
        if self._eqattr is None:
            # Compare object
            return id(other) == id(self)
        else:
            # Compare defined attribute
            return other == getattr(self, self._eqattr)

    def __ne__(self, other):
        """Comparison method: this object is treated like the attribute
           defined by set_eqattr()
        """
        return not self.__eq__(other)

    def __repr__(self):
        """String representation of object

           The representation depends on the verbose level set by debug_repr().
           If set to 0 the generic object representation is returned, else
           the representation of the object includes all object attributes
           and their values with proper indentation.
        """
        return self._str_repr(True)

    def __str__(self):
        """Informal string representation of object

           The representation depends on the verbose level set by debug_repr().
           If set to 0 the generic object representation is returned, else
           the representation of the object includes all object attributes
           and their values.
        """
        return self._str_repr()

    def _str_repr(self, isrepr=False):
        """String representation of object"""
        global _rlevel
        if _rlevel == 0:
            # Return generic object representation
            if isrepr:
                return super(BaseObj, self).__repr__()
            else:
                return super(BaseObj, self).__str__()
        elif not isrepr:
            if _rlevel == 1 and self._strfmt1 is not None:
                return self.format(self._strfmt1)
            elif _rlevel == 2 and self._strfmt2 is not None:
                return self.format(self._strfmt2)

        # Representation of object with proper indentation
        out = []
        if self._attrlist is None:
            attrlist = sorted(self.__dict__.keys())
        else:
            attrlist = self._attrlist
        for key in attrlist:
            if key[0] != '_':
                val = getattr(self, key, None)
                if val != None:
                    if isrepr:
                        value = pformat(val, indent=0)
                        if (isinstance(val, list) or isinstance(val, dict)) and value.find("\n") > 0:
                            # If list or dictionary have more than one line as
                            # returned from pformat, add an extra new line
                            # between opening and closing brackets and add
                            # another indentation to the body
                            value = (value[0] + "\n" + value[1:-1]).replace("\n", "\n"+_sindent) + "\n" + value[-1]
                        out.append("%s%s = %s,\n" % (_sindent, key, value.replace("\n", "\n"+_sindent)))
                    else:
                        out.append("%s=%s" % (key, self._str_value(val)))
        name = self.__class__.__name__
        if isrepr:
            joinstr = ""
            if len(out) > 0:
                out.insert(0, "\n")
        else:
            joinstr = ", "
        return "%s(%s)" % (name, joinstr.join(out))

    def _str_value(self, value):
        """Format value"""
        if isinstance(value, list) or isinstance(value, tuple):
            # Display list or tuple
            out = []
            for item in value:
                out.append(self._str_value(item))
            return '[' + ', '.join(out) + ']'
        elif isinstance(value, dict):
            # Display dictionary
            out = []
            for key,val in value.iteritems():
                out.append(str(key) + ": " + self._str_value(val))
            return '{' + ', '.join(out) + '}'
        elif type(value) == int or type(value) == long or type(value) == str:
            if _strsize > 0 and type(value) == str:
                return repr(value[:_strsize])
            return repr(value)
        else:
            return str(value)

    def set_attrlist(self, attr):
        """Add list of attribute names in object to display by str() or repr()

           attr:
               Name or list of names to add to the list of attribute names
               to display
        """
        if self._attrlist is None:
            self._attrlist = []
        if isinstance(attr, list):
            # Add given list of items
            self._attrlist += attr
        else:
            # Add a single item
            self._attrlist.append(attr)

    def set_attr(self, name, value, switch=False):
        """Add name/value as an object attribute and add the name to the
           list of attributes to display

           name:
               Attribute name
           value:
               Attribute value
        """
        setattr(self, name, value)
        self.set_attrlist(name)
        if switch:
            if self._attrs is None:
                self._attrs = {}
            # Make a reference to name
            self._attrs["switch"] = name
            if self._fattrs is None:
                self._fattrs = []
            # Make it a flat object
            self._fattrs.append(name)

    def set_eqattr(self, attr):
        """Set the comparison attribute

           attr:
               Attribute to use for object comparison

           Examples:
               x = BaseObj(a=1, b=2)
               x.set_eqattr("a")
               x == 1 will return True, the same as x.a == 1
        """
        self._eqattr = attr

    def set_strfmt(self, level, format):
        """Save format for given display level

           level:
               Display level given as a first argument
           format:
               String format for given display level, given as a second argument
        """
        if level == 1:
            self._strfmt1 = format
        elif level == 2:
            self._strfmt2 = format
        else:
            raise Exception("Invalid string format level [%d]" % level)

    def set_global(self, name, value):
        """Set global variable."""
        self._globals[name] = value

    @staticmethod
    def debug_repr(level=None):
        """Return or set verbose level of object's string representation.
           When setting the verbose level, return the verbose level before
           setting it.

           level:
               Level of verbosity to set

           Examples:
               # Set verbose level to its minimal object representation
               x.debug_repr(0)

               # Object representation is a bit more verbose
               x.debug_repr(1)

               # Object representation is a lot more verbose
               x.debug_repr(2)
        """
        global _rlevel
        ret = _rlevel
        if level is not None:
            _rlevel = level
        return ret

    def debug_level(self, level=0):
        """Set debug level mask.

           level:
               Level to set. This could be a number or a string expression
               of names defined by debug_map()

           Examples:
               # Set level
               x.debug_level(0xFF)

               # Set level using expression
               x.debug_level('all')
               x.debug_level('debug ^ 1')
        """
        global _dlevel
        if type(level) == str:
            # Convert named verbose levels to a number
            # -- Get a list of all named verbose levels
            for item in sorted(set(re.split('\W+', level))):
                if len(item) > 0:
                    if item in _debug_map:
                        # Replace all occurrences of named verbose level
                        # to its corresponding numeric value
                        level = re.sub(r'\b' + item + r'\b', hex(_debug_map[item]), level)
                    else:
                        try:
                            # Find out if verbose is a number
                            # (decimal, hex, octal, ...)
                            tmp = int(item, 0)
                        except:
                            raise Exception("Unknown debug level [%s]" % item)
            # Evaluate the whole expression
            _dlevel = eval(level)
        else:
            # Already a number
            _dlevel = level
        return _dlevel

    @staticmethod
    def debug_map(bitmap, name='', disp=''):
        """Add a debug mapping.

           Generic debug levels map
             <bitmap>  <name>  <disp prefix>
              0x000    'none'
              0x001    'info'  'INFO: ' # Display info messages only
              0x0FF    'debug' 'DBG:  ' # Display info and all debug messages (0x02-0x80)
             >0x100    user defined verbose levels
        """
        if name:
            _debug_map[name] = bitmap
        if disp:
            _debug_prefix[bitmap] = disp

    @staticmethod
    def dindent(indent=None):
        """Set global dprint indentation."""
        global _dindent
        if indent is not None:
            _dindent = " " * indent
        return _dindent

    @staticmethod
    def sindent(indent=None):
        """Set global object indentation."""
        global _sindent
        if indent is not None:
            _sindent = " " * indent
        return _sindent

    @staticmethod
    def strsize(size):
        """Set global string truncation."""
        global _strsize
        _strsize = size

    @staticmethod
    def tstamp(enable=None, fmt=None):
        """Enable/disable timestamps on dprint messages and/or
           set the default format for timestamps

           enable:
               Boolean to enable/disable timestamps
           fmt:
               Set timestamp format
        """
        global _tstamp,_tstampfmt
        if enable is not None:
            _tstamp = enable
        if fmt is not None:
            _tstampfmt = fmt

    @staticmethod
    def timestamp(fmt=None):
        """Return the timestamp if it is enabled.

           fmt:
               Timestamp format, default is given by the format
               set by tstamp()
        """
        if _tstamp:
            if fmt is None:
                fmt = _tstampfmt
            return fstrobj.format(fmt, time.time())
        return ""

    def open_log(self, logfile):
        """Open log file."""
        global _logfh
        self.close_log()
        _logfh = open(logfile, "w")

    def close_log(self):
        """Close log file."""
        global _logfh
        if _logfh != None:
            _logfh.close()
            _logfh = None

    @staticmethod
    def write_log(data):
        """Write data to log file."""
        if _logfh != None:
            _logfh.write(data + "\n")

    @staticmethod
    def flush_log():
        """Flush data to log file."""
        if _logfh != None:
            _logfh.flush()

    @staticmethod
    def dprint_count():
        """Return the number of dprint messages actually displayed."""
        return _dcount

    def format(self, fmt, *kwts, **kwds):
        """Format the arguments and return the string using the format given.
           If no arguments are given either positional or named then object
           attributes set by set_attrlist() are used as positional arguments
           and all object attributes are used as named arguments

           fmt:
               String format to use for the arguments, where {0}, {1}, etc.
               are used for positional arguments and {name1}, {name2}, etc.
               are used for named arguments given after fmt.
        """
        if len(kwts) == 0 and len(kwds) == 0:
            # Use object attributes, both positional using _attrlist and
            # named arguments using object's own dictionary
            if self._attrlist is not None:
                kwts = (getattr(self, attr) for attr in self._attrlist)
            kwds = self.__dict__.copy()
            if self._globals:
                # Include the shared attributes as named attributes
                kwds.update(self._globals)
        return fstrobj.format(fmt, *kwts, **kwds)

    def dprint(self, level, msg, indent=0):
        """Print debug message if level is allowed by the verbose level
           given in debug_level().
        """
        ret = ''
        if level is None:
            return
        if type(level) == str:
            level = _debug_map[level.lower()]
        if level & _dlevel:
            # Add display prefix only if msg is not an empty string
            if len(msg):
                # Find the right display prefix
                prefix = _dindent
                for bitmap in sorted(_debug_prefix):
                    if level & bitmap:
                        prefix += _debug_prefix[bitmap]
                        break
                # Add display prefix to the message
                ret = prefix + self.timestamp()
                if indent > 0:
                    ret += " " * indent
                ret += msg
                indent += len(prefix)
            if indent > 0:
                sp = ' ' * indent
                ret = ret.replace("\n", "\n"+sp)
            print ret
            self.write_log(ret)
            global _dcount
            _dcount += 1



#!/usr/bin/env python
#===============================================================================
# Copyright 2012 NetApp, Inc. All Rights Reserved,
# contribution by Jorge Mora <mora@netapp.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#===============================================================================
#
# To create man pages:
# $ python setup.py build
#
# To install, run as root:
# $ python setup.py install
#
# To create an rpm (need to create the man pages first):
# $ python setup.py build
# $ python setup.py bdist_rpm --release p2.6
#
import os
import nfstest_config as c
from tools import create_manpage
from distutils.core import setup
from distutils.command.build import build

class Build(build):
    def run(self):
        create_manpage.run()
        build.run(self)

setup(
    name             = c.NFSTEST_PACKAGE,
    version          = c.NFSTEST_VERSION,
    description      = c.NFSTEST_SUMMARY,
    long_description = c.NFSTEST_DESCRIPTION,
    author           = c.NFSTEST_AUTHOR,
    author_email     = c.NFSTEST_AUTHOR_EMAIL,
    maintainer       = c.NFSTEST_MAINTAINER,
    maintainer_email = c.NFSTEST_MAINTAINER_EMAIL,
    license          = c.NFSTEST_LICENSE,
    url              = c.NFSTEST_URL,
    download_url     = c.NFSTEST_DL_URL,
    py_modules       = c.NFSTEST_MODULES,
    packages         = c.NFSTEST_PACKAGES,
    scripts          = c.NFSTEST_SCRIPTS,
    cmdclass = {'build': Build},
    data_files = [
        # Man pages for scripts
        (os.path.join(c.NFSTEST_USRMAN, 'man1'), c.NFSTEST_MAN1),
        (os.path.join(c.NFSTEST_USRMAN, 'man3'), c.NFSTEST_MAN3),
    ],
)



#===============================================================================
# Copyright 2014 NetApp, Inc. All Rights Reserved,
# contribution by Jorge Mora <mora@netapp.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#===============================================================================
# Generated by process_xdr.py from mount3.x on Sun Oct 04 08:09:27 2015
"""
MOUNTv3 constants module
"""
import nfstest_config as c

# Module constants
__author__    = "Jorge Mora (%s)" % c.NFSTEST_AUTHOR_EMAIL
__copyright__ = "Copyright (C) 2014 NetApp, Inc."
__license__   = "GPL v2"
__version__   = "3.0"

#
# Sizes
MNTPATHLEN = 1024  # Maximum bytes in a path name
MNTNAMLEN  = 255   # Maximum bytes in a name
FHSIZE3    = 64    # Maximum bytes in a V3 file handle

# Enum mountstat3
MNT3_OK             = 0      # no error
MNT3ERR_PERM        = 1      # Not owner
MNT3ERR_NOENT       = 2      # No such file or directory
MNT3ERR_IO          = 5      # I/O error
MNT3ERR_ACCES       = 13     # Permission denied
MNT3ERR_NOTDIR      = 20     # Not a directory
MNT3ERR_INVAL       = 22     # Invalid argument
MNT3ERR_NAMETOOLONG = 63     # Filename too long
MNT3ERR_NOTSUPP     = 10004  # Operation not supported
MNT3ERR_SERVERFAULT = 10006  # A failure on the server

mountstat3 = {
        0 : "MNT3_OK",
        1 : "MNT3ERR_PERM",
        2 : "MNT3ERR_NOENT",
        5 : "MNT3ERR_IO",
       13 : "MNT3ERR_ACCES",
       20 : "MNT3ERR_NOTDIR",
       22 : "MNT3ERR_INVAL",
       63 : "MNT3ERR_NAMETOOLONG",
    10004 : "MNT3ERR_NOTSUPP",
    10006 : "MNT3ERR_SERVERFAULT",
}

# Enum rpc_auth_flavors
AUTH_NULL      = 0
AUTH_UNIX      = 1
AUTH_SHORT     = 2
AUTH_DES       = 3
AUTH_KRB       = 4
AUTH_GSS       = 6
AUTH_MAXFLAVOR = 8
# pseudoflavors:
AUTH_GSS_KRB5  = 390003
AUTH_GSS_KRB5I = 390004
AUTH_GSS_KRB5P = 390005
AUTH_GSS_LKEY  = 390006
AUTH_GSS_LKEYI = 390007
AUTH_GSS_LKEYP = 390008
AUTH_GSS_SPKM  = 390009
AUTH_GSS_SPKMI = 390010
AUTH_GSS_SPKMP = 390011

rpc_auth_flavors = {
         0 : "AUTH_NULL",
         1 : "AUTH_UNIX",
         2 : "AUTH_SHORT",
         3 : "AUTH_DES",
         4 : "AUTH_KRB",
         6 : "AUTH_GSS",
         8 : "AUTH_MAXFLAVOR",
    390003 : "AUTH_GSS_KRB5",
    390004 : "AUTH_GSS_KRB5I",
    390005 : "AUTH_GSS_KRB5P",
    390006 : "AUTH_GSS_LKEY",
    390007 : "AUTH_GSS_LKEYI",
    390008 : "AUTH_GSS_LKEYP",
    390009 : "AUTH_GSS_SPKM",
    390010 : "AUTH_GSS_SPKMI",
    390011 : "AUTH_GSS_SPKMP",
}

# Enum mount_proc3
MOUNTPROC3_NULL    = 0
MOUNTPROC3_MNT     = 1
MOUNTPROC3_DUMP    = 2
MOUNTPROC3_UMNT    = 3
MOUNTPROC3_UMNTALL = 4
MOUNTPROC3_EXPORT  = 5

mount_proc3 = {
    0 : "MOUNTPROC3_NULL",
    1 : "MOUNTPROC3_MNT",
    2 : "MOUNTPROC3_DUMP",
    3 : "MOUNTPROC3_UMNT",
    4 : "MOUNTPROC3_UMNTALL",
    5 : "MOUNTPROC3_EXPORT",
}



#===============================================================================
# Copyright 2015 NetApp, Inc. All Rights Reserved,
# contribution by Jorge Mora <mora@netapp.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#===============================================================================
"""
KRB5 module

Decode KRB5 layer
Decoding using ASN.1 DER (Distinguished Encoding Representation)

RFC 4120 The Kerberos Network Authentication Service (V5)
RFC 6113 A Generalized Framework for Kerberos Pre-Authentication
"""
from packet.utils import *
import krb5_const as const
import nfstest_config as c
from baseobj import BaseObj
from packet.derunpack import DERunpack

# Module constants
__author__    = "Jorge Mora (%s)" % c.NFSTEST_AUTHOR_EMAIL
__copyright__ = "Copyright (C) 2015 NetApp, Inc."
__license__   = "GPL v2"
__version__   = "1.0"

def SequenceOf(obj, objtype):
    """SEQUENCE OF: return list of the given object type"""
    ret = []
    if obj is not None:
        for item in obj:
            ret.append(objtype(item))
    return ret

def Optional(obj, objtype):
    """Get Optional item of the given object type"""
    if obj is not None:
        return objtype(obj)

def KerberosTime(stime, usec=None):
    """Convert floating point time to a DateStr object,
       include the microseconds if given
    """
    if stime is not None:
        if usec is not None:
            stime += (0.000001*usec)
        return DateStr(stime)

class KDCOptions(OptionFlags):
    """KDC Option flags"""
    _bitnames = const.kdc_options
    _reversed = 31

class APOptions(OptionFlags):
    """AP Option flags"""
    _bitnames = const.ap_options
    _reversed = 31

# Application Tag Numbers
class krb5_application(Enum):
    """enum krb5_application"""
    _enumdict = const.krb5_application

# Principal Names
class krb5_principal(Enum):
    """enum krb5_principal"""
    _enumdict = const.krb5_principal

# Pre-authentication and Typed Data
class krb5_patype(Enum):
    """enum krb5_patype"""
    _enumdict = const.krb5_patype

# Address Types
class krb5_addrtype(Enum):
    """enum krb5_addrtype"""
    _enumdict = const.krb5_addrtype

# Authorization Data Types
class krb5_adtype(Enum):
    """enum krb5_adtype"""
    _enumdict = const.krb5_adtype

# Kerberos Encryption Type Numbers
class krb5_etype(Enum):
    """enum krb5_etype"""
    _enumdict = const.krb5_etype

# Kerberos Checksum Type Numbers
class krb5_ctype(Enum):
    """enum krb5_ctype"""
    _enumdict = const.krb5_ctype

# Kerberos Fast Armor Type Numbers
class krb5_fatype(Enum):
    """enum krb5_fatype"""
    _enumdict = const.krb5_fatype

# Error Codes
class krb5_status(Enum):
    """enum krb5_status"""
    _enumdict = const.krb5_status

class PrincipalName(BaseObj):
    """
       PrincipalName  ::= SEQUENCE {
           name-type    [0] Int32,
           name-string  [1] SEQUENCE OF KerberosString
       }
    """
    # Class attributes
    _strfmt1  = "{0} {1}"
    _attrlist = ("ntype", "name")

    def __init__(self, obj):
        self.ntype = krb5_principal(obj.get(0))
        self.name  = obj.get(1)

class HostAddress(BaseObj):
    """
       HostAddress  ::= SEQUENCE  {
           addr-type  [0] Int32,
           address    [1] OCTET STRING
       }
    """
    # Class attributes
    _strfmt1  = "{0} {1}"
    _attrlist = ("atype", "address")

    def __init__(self, obj):
        self.atype   = krb5_addrtype(obj.get(0))
        self.address = obj.get(1)

class EtypeInfo2Entry(BaseObj):
    """
       ETYPE-INFO2-ENTRY  ::= SEQUENCE {
           etype      [0] Int32,
           salt       [1] KerberosString OPTIONAL,
           s2kparams  [2] OCTET STRING OPTIONAL
       }
    """
    # Class attributes
    _attrlist = ("etype", "salt", "s2kparams")

    def __init__(self, obj):
        self.etype     = krb5_etype(obj.get(0))
        self.salt      = obj.get(1)
        self.s2kparams = obj.get(2)

class Checksum(BaseObj):
    """
       Checksum  ::= SEQUENCE {
           cksumtype  [0] Int32,
           checksum   [1] OCTET STRING
       }
    """
    # Class attributes
    _strfmt2  = "Checksum(ctype={0})"
    _attrlist = ("ctype", "checksum")

    def __init__(self, obj):
        self.ctype    = krb5_ctype(obj.get(0))
        self.checksum = obj.get(1)

class KrbFastArmor(BaseObj):
    """
       KrbFastArmor  ::= SEQUENCE {
           armor-type   [0] Int32,
               -- Type of the armor.
           armor-value  [1] OCTET STRING,
               -- Value of the armor.
       }
    """
    # Class attributes
    _strfmt2  = "KrbFastArmor(fatype={0})"
    _attrlist = ("fatype", "value")

    def __init__(self, obj):
        self.fatype = krb5_fatype(obj.get(0))
        self.value  = obj.get(1)

class KrbFastArmoredReq(BaseObj):
    """
       KrbFastArmoredReq ::= SEQUENCE {
           armor        [0] KrbFastArmor OPTIONAL,
               -- Contains the armor that identifies the armor key.
               -- MUST be present in AS-REQ.
           req-checksum [1] Checksum,
               -- For AS, contains the checksum performed over the type
               -- KDC-REQ-BODY for the req-body field of the KDC-REQ
               -- structure;
               -- For TGS, contains the checksum performed over the type
               -- AP-REQ in the PA-TGS-REQ padata.
               -- The checksum key is the armor key, the checksum
               -- type is the required checksum type for the enctype of
               -- the armor key, and the key usage number is
               -- KEY_USAGE_FAST_REQ_CHKSUM.
           enc-fast-req [2] EncryptedData, -- KrbFastReq --
               -- The encryption key is the armor key, and the key usage
               -- number is KEY_USAGE_FAST_ENC.
       }
    """
    # Class attributes
    _attrlist = ("armor", "checksum", "enc_fast")

    def __init__(self, obj):
        self.armor    = Optional(obj.get(0), KrbFastArmor)
        self.checksum = Checksum(obj.get(1))
        self.enc_fast = EncryptedData(obj.get(2))

class KrbFastArmoredRep(BaseObj):
    """
       KrbFastArmoredRep ::= SEQUENCE {
          enc-fast-rep  [0] EncryptedData, -- KrbFastResponse --
              -- The encryption key is the armor key in the request, and
              -- the key usage number is KEY_USAGE_FAST_REP.
       }
    """
    # Class attributes
    _attrlist = ("enc_fast",)

    def __init__(self, obj):
        self.enc_fast = EncryptedData(obj.get(0))

class paData(BaseObj):
    """
       PA-DATA  ::= SEQUENCE {
           -- NOTE: first tag is [1], not [0]
           padata-type   [1] Int32,
           padata-value  [2] OCTET STRING
       }
    """
    # Class attributes
    _attrlist = ("patype", "value")

    def __init__(self, obj):
        self.patype = krb5_patype(obj.get(1))
        self.value  = obj.get(2)

        if len(self.value) > 0:
            if self.patype == const.PA_ETYPE_INFO2:
                self.value = SequenceOf(DERunpack(self.value).get_item(), EtypeInfo2Entry)
            elif self.patype == const.PA_ENC_TIMESTAMP:
                self.value = EncryptedData(DERunpack(self.value).get_item())
            elif self.patype == const.PA_TGS_REQ:
                self.value = AP_REQ(DERunpack(self.value).get_item())
            elif self.patype == const.PA_FX_FAST:
                pobj = DERunpack(self.value).get_item()
                # Get the CHOICE tag and value
                tag, value = pobj.popitem()
                if tag == 0:
                    if len(value) == 1:
                        # PA-FX-FAST-REPLY ::= CHOICE {
                        #     armored-data [0] KrbFastArmoredRep,
                        # }
                        self.value = KrbFastArmoredRep(value)
                    else:
                        # PA-FX-FAST-REQUEST ::= CHOICE {
                        #     armored-data [0] KrbFastArmoredReq,
                        # }
                        self.value = KrbFastArmoredReq(value)

class EncryptedData(BaseObj):
    """
       EncryptedData  ::= SEQUENCE {
           etype   [0] Int32 -- EncryptionType --,
           kvno    [1] UInt32 OPTIONAL,
           cipher  [2] OCTET STRING -- ciphertext
       }
    """
    # Class attributes
    _strfmt2  = "EncryptedData(etype={0})"
    _attrlist = ("etype", "kvno", "cipher")

    def __init__(self, obj):
        self.etype  = krb5_etype(obj.get(0))
        self.kvno   = obj.get(1)
        self.cipher = obj.get(2)

class Ticket(BaseObj):
    """
       Ticket  ::= [APPLICATION 1] SEQUENCE {
           tkt-vno   [0] INTEGER (5),
           realm     [1] Realm,
           sname     [2] PrincipalName,
           enc-part  [3] EncryptedData -- EncTicketPart
       }
    """
    # Class attributes
    _attrlist = ("tkt_vno", "realm", "sname", "enc_part")

    def __init__(self, obj):
        obj = obj[1] # Application 1
        self.tkt_vno  = obj.get(0)
        self.realm    = obj.get(1)
        self.sname    = PrincipalName(obj.get(2))
        self.enc_part = EncryptedData(obj.get(3))

class AP_REQ(BaseObj):
    """
       AP-REQ  ::= [APPLICATION 14] SEQUENCE {
           pvno           [0] INTEGER (5),
           msg-type       [1] INTEGER (14),
           options        [2] APOptions,
           ticket         [3] Ticket,
           authenticator  [4] EncryptedData -- Authenticator
       }
    """
    # Class attributes
    _attrlist = ("pvno", "msgtype", "options", "ticket", "authenticator")

    def __init__(self, obj):
        obj = obj[14] # Application 14
        self.pvno          = obj.get(0)
        self.msgtype       = krb5_application(obj.get(1))
        self.options       = APOptions(obj.get(2))
        self.ticket        = Ticket(obj.get(3))
        self.authenticator = EncryptedData(obj.get(4))

class AP_REP(BaseObj):
    """
       AP-REP  ::= [APPLICATION 15] SEQUENCE {
           pvno      [0] INTEGER (5),
           msg-type  [1] INTEGER (15),
           enc-part  [2] EncryptedData -- EncAPRepPart
       }
    """
    # Class attributes
    _attrlist = ("pvno", "msgtype", "enc_part")

    def __init__(self, obj):
        obj = obj[15] # Application 15
        self.pvno     = obj.get(0)
        self.msgtype  = krb5_application(obj.get(1))
        self.enc_part = EncryptedData(obj.get(2))

class KDC_REQ_BODY(BaseObj):
    """
       KDC-REQ-BODY  ::= SEQUENCE {
           options                  [0] KDCOptions,
           cname                    [1] PrincipalName OPTIONAL
                                        -- Used only in AS-REQ --,
           realm                    [2] Realm
                                        -- Server's realm
                                        -- Also client's in AS-REQ --,
           sname                    [3] PrincipalName OPTIONAL,
           from                     [4] KerberosTime OPTIONAL,
           till                     [5] KerberosTime,
           rtime                    [6] KerberosTime OPTIONAL,
           nonce                    [7] UInt32,
           etype                    [8] SEQUENCE OF Int32 -- EncryptionType
                                        -- in preference order --,
           addresses                [9] HostAddresses OPTIONAL,
           enc-authorization-data  [10] EncryptedData OPTIONAL
                                        -- AuthorizationData --,
           additional-tickets      [11] SEQUENCE OF Ticket OPTIONAL
                                        -- NOTE: not empty
       }
    """
    # Class attributes
    _attrlist = ("options", "cname", "realm", "sname", "stime", "etime",
                 "rtime", "nonce", "etype", "addrs", "edata", "tickets")

    def __init__(self, obj):
        self.options = KDCOptions(obj.get(0))
        self.cname   = Optional(obj.get(1), PrincipalName)
        self.realm   = obj.get(2)
        self.sname   = Optional(obj.get(3), PrincipalName)
        self.stime   = KerberosTime(obj.get(4))
        self.etime   = KerberosTime(obj.get(5))
        self.rtime   = KerberosTime(obj.get(6))
        self.nonce   = obj.get(7)
        self.etype   = [krb5_etype(x) for x in obj.get(8)]
        self.addrs   = SequenceOf(obj.get(9), HostAddress)
        self.edata   = Optional(obj.get(10), EncryptedData)
        self.tickets = SequenceOf(obj.get(11), Ticket)

        if self.cname is not None:
            self.set_strfmt(1, "cname:({1}) {2}")
        else:
            self.set_strfmt(1, "sname:({3}) {2}")

class KDC_REQ(BaseObj):
    """
       KDC-REQ  ::= SEQUENCE {
           -- NOTE: first tag is [1], not [0]
           pvno      [1] INTEGER (5) ,
           msg-type  [2] INTEGER (10 -- AS -- | 12 -- TGS --),
           padata    [3] SEQUENCE OF PA-DATA OPTIONAL
                         -- NOTE: not empty --,
           req-body  [4] KDC-REQ-BODY
       }
    """
    # Class attributes
    _strfmt1  = "KRB{0} {1} {3}"
    _attrlist = ("pvno", "msgtype", "padata", "body")

    def __init__(self, obj):
        self.pvno    = obj.get(1)
        self.msgtype = krb5_application(obj.get(2))
        self.padata  = SequenceOf(obj.get(3), paData)
        self.body    = KDC_REQ_BODY(obj.get(4))

class KDC_REP(BaseObj):
    """
       KDC-REP  ::= SEQUENCE {
           pvno      [0] INTEGER (5),
           msg-type  [1] INTEGER (11 -- AS -- | 13 -- TGS --),
           padata    [2] SEQUENCE OF PA-DATA OPTIONAL
                         -- NOTE: not empty --,
           crealm    [3] Realm,
           cname     [4] PrincipalName,
           ticket    [5] Ticket,
           enc-part  [6] EncryptedData
                         -- EncASRepPart or EncTGSRepPart,
                         -- as appropriate
       }
    """
    # Class attributes
    _strfmt1  = "KRB{0} {1} cname:({4}) {3}"
    _attrlist = ("pvno", "msgtype", "padata", "crealm", "cname",
                 "ticket", "enc_part")

    def __init__(self, obj):
        self.pvno     = obj.get(0)
        self.msgtype  = krb5_application(obj.get(1))
        self.padata   = SequenceOf(obj.get(2), paData)
        self.crealm   = obj.get(3)
        self.cname    = PrincipalName(obj.get(4))
        self.ticket   = Ticket(obj.get(5))
        self.enc_part = EncryptedData(obj.get(6))

class KRB_ERROR(BaseObj):
    """
       KRB-ERROR  ::= [APPLICATION 30] SEQUENCE {
           pvno        [0] INTEGER (5),
           msg-type    [1] INTEGER (30),
           ctime       [2] KerberosTime OPTIONAL,
           cusec       [3] Microseconds OPTIONAL,
           stime       [4] KerberosTime,
           susec       [5] Microseconds,
           error-code  [6] Int32,
           crealm      [7] Realm OPTIONAL,
           cname       [8] PrincipalName OPTIONAL,
           realm       [9] Realm -- service realm --,
           sname       [10] PrincipalName -- service name --,
           e-text      [11] KerberosString OPTIONAL,
           e-data      [12] OCTET STRING OPTIONAL
       }
    """
    # Class attributes
    _strfmt1  = "KRB{0} {4}"
    _attrlist = ("pvno", "msgtype", "ctime", "stime", "error", "crealm",
                 "cname", "realm", "sname", "etext", "edata")

    def __init__(self, obj):
        # Application 30: do not process the application here, it should be
        # done at the parent class to know what type of object to instantiate
        self.pvno    = obj.get(0)
        self.msgtype = krb5_application(obj.get(1))
        self.ctime   = KerberosTime(obj.get(2), obj.get(3))
        self.stime   = KerberosTime(obj.get(4), obj.get(5))
        self.error   = krb5_status(obj.get(6))
        self.crealm  = obj.get(7)
        self.cname   = Optional(obj.get(8), PrincipalName)
        self.realm   = obj.get(9)
        self.sname   = PrincipalName(obj.get(10))
        self.etext   = obj.get(11)
        edata        = obj.get(12)
        if edata is not None:
            if self.error == const.KDC_ERR_PREAUTH_REQUIRED:
                edata = SequenceOf(DERunpack(edata).get_item(), paData)
        self.edata = edata

class KRB5(BaseObj):
    """KRB5 object

       Usage:
           from packet.application.krb5 import KRB5

           # Decode KRB5 layer
           x = KRB5(pktt, proto)

       Object definition:

       KRB5(
           appid = int,  # Application Identifier
           kdata = KDC_REQ|KDC_REP|KRB_ERROR
       }
    """
    # Class attributes
    _fattrs   = ("kdata",)
    _strfmt1  = "{1}"
    _attrlist = ("appid", "kdata")

    def __init__(self, pktt, proto):
        """Constructor

           Initialize object's private data.

           pktt:
               Packet trace object (packet.pktt.Pktt) so this layer has
               access to the parent layers.
           proto:
               Transport layer protocol.
        """
        self._krb5 = False

        try:
            unpack = pktt.unpack
            if proto == 6:
                # Get the length of the TCP record
                length = unpack.unpack_ushort()
                if length < len(unpack):
                    return

            slen = unpack.size()
            derunpack = DERunpack(unpack.getbytes())
            krbobj = derunpack.get_item()
            appid, obj = krbobj.items()[0]
            self.appid = krb5_application(appid)
            if self.appid in (const.AS_REQ, const.TGS_REQ):
                # AS-REQ  ::= [APPLICATION 10] KDC-REQ
                # TGS-REQ ::= [APPLICATION 12] KDC-REQ
                self.kdata = KDC_REQ(obj)
            elif self.appid in (const.AS_REP, const.TGS_REP):
                # AS-REP  ::= [APPLICATION 11] KDC-REP
                # TGS-REP ::= [APPLICATION 13] KDC-REP
                self.kdata = KDC_REP(obj)
            elif self.appid == const.KRB_ERROR:
                self.kdata = KRB_ERROR(obj)
            else:
                self.kdata = obj
        except Exception as e:
            return
        if len(derunpack) > 0:
            return
        unpack.read(slen)
        self._krb5 = True

    def __nonzero__(self):
        """Truth value testing for the built-in operation bool()"""
        return self._krb5



#===============================================================================
# Copyright 2013 NetApp, Inc. All Rights Reserved,
# contribution by Jorge Mora <mora@netapp.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#===============================================================================
"""
GSS module

Decode GSS layers.

RFC 2203 RPCSEC_GSS Protocol Specification
RFC 1964 The Kerberos Version 5 GSS-API Mechanism

NOTE:
  Only procedures RPCSEC_GSS_INIT and RPCSEC_GSS_DATA are supported
"""
import krb5
import rpc_const
import gss_const as const
from packet.utils import *
import nfstest_config as c
from baseobj import BaseObj
from packet.derunpack import DERunpack

# Module constants
__author__    = "Jorge Mora (%s)" % c.NFSTEST_AUTHOR_EMAIL
__copyright__ = "Copyright (C) 2013 NetApp, Inc."
__license__   = "GPL v2"
__version__   = "1.4"

# Token Identifier TOK_ID
KRB_AP_REQ       = 0x0100
KRB_AP_REP       = 0x0200
KRB_ERROR        = 0x0300
KRB_TOKEN_GETMIC = 0x0101

# Integrity algorithm indicator
class gss_sgn_alg(Enum):
    """enum gss_sgn_alg"""
    _enumdict = const.gss_sgn_alg

# GSS Major Status Codes
class gss_major_status(Enum):
    """enum gss_major_status"""
    _enumdict = const.gss_major_status

# GSS Minor Status Codes
class gss_minor_status(Enum):
    """enum gss_minor_status"""
    _enumdict = const.gss_minor_status

class GetMIC(BaseObj):
    """struct GSS_GetMIC {
           unsigned short sgn_alg;      /* Integrity algorithm indicator */
           opaque         filler[4];    /* Filler bytes: 0xffffffff */
           opaque         snd_seq[8];   /* Sequence number field */
           opaque         sgn_cksum[8]; /* Checksum of "to-be-signed data" */
       };
    """
    # Class attributes
    _strfmt2  = "GetMIC({0}, snd_seq:{1}, sgn_cksum:{2})"
    _attrlist = ("sgn_alg", "snd_seq", "sgn_cksum")

    def __init__(self, unpack):
        ulist = unpack.unpack(22, "!H4s8s8s")
        self.sgn_alg   = gss_sgn_alg(ulist[0])
        self.filler    = ulist[1]
        self.snd_seq   = StrHex(ulist[2])
        self.sgn_cksum = StrHex(ulist[3])

class GSS_API(BaseObj):
    """GSS-API DEFINITIONS ::=

       BEGIN

       MechType ::= OBJECT IDENTIFIER
       -- representing Kerberos V5 mechanism

       GSSAPI-Token ::=
       -- option indication (delegation, etc.) indicated within
       -- mechanism-specific token
       [APPLICATION 0] IMPLICIT SEQUENCE {
               thisMech MechType,
               innerToken ANY DEFINED BY thisMech
                  -- contents mechanism-specific
                  -- ASN.1 structure not required
               }

       END
    """
    # Class attributes
    _strfmt2  = "GSS_API(OID:{0}, {2})"
    _attrlist = ("oid", "tok_id", "krb5")

    def __init__(self, data):
        self._valid = False
        derunpack = DERunpack(data)
        # Get the Kerberos 5 OID only -- from application 0
        krbobj = derunpack.get_item(oid="1.2.840.113554.1.2.2").get(0)
        if krbobj is not None and len(krbobj) > 0:
            self.oid    = krbobj.get(0)
            self.tok_id = derunpack.unpack_ushort()
            self.krb5   = None

            if self.tok_id == KRB_AP_REQ:
                krbobj = derunpack.get_item()
                self.krb5 = krb5.AP_REQ(krbobj)
                self.krb5.set_strfmt(2, "{1}, opts:{2}, Ticket({3})")
                self.krb5.ticket.set_strfmt(2, "{2}@{1}({2.ntype}), {3.etype}")
                self.krb5.ticket.sname.set_strfmt(2, "{1:/:}")
            elif self.tok_id == KRB_AP_REP:
                krbobj = derunpack.get_item()
                self.krb5 = krb5.AP_REP(krbobj)
                self.krb5.set_strfmt(2, "{1}, {2.etype}")
            elif self.tok_id == KRB_ERROR:
                krbobj = derunpack.get_item()
                self.krb5 = krb5.KRB_ERROR(krbobj.get(30))
            elif self.tok_id == KRB_TOKEN_GETMIC:
                self.krb5 = GetMIC(derunpack)
            else:
                self.krb5 = StrHex(derunpack.getbytes())
            self._valid = True

    def __nonzero__(self):
        """Truth value testing for the built-in operation bool()"""
        return self._valid

class GSS_init_arg(BaseObj):
    """struct rpc_gss_init_arg {
           opaque  gss_token<>;
       };
    """
    # Class attributes
    _strfmt2  = "token: {0:#x:.32}..."
    _attrlist = ("token",)

    def __init__(self, unpack):
        self.token = unpack.unpack_opaque()
        krb5 = GSS_API(self.token)
        if krb5:
            self.token = krb5
            self.set_strfmt(2, "{0}")

class GSS_init_res(BaseObj):
    """struct rpc_gss_init_res {
           opaque        handle<>;
           unsigned int  gss_major;
           unsigned int  gss_minor;
           unsigned int  seq_window;
           opaque        gss_token<>;
       };
    """
    # Class attributes
    _strfmt2  = "context: {0}, token: {4:#x:.16}..."
    _attrlist = ("context", "major", "minor", "seq_window", "token")

    def __init__(self, unpack):
        self.context    = StrHex(unpack.unpack_opaque())
        self.major      = gss_major_status(unpack.unpack_uint())
        self.minor      = gss_minor_status(unpack.unpack_int())
        self.seq_window = unpack.unpack_uint()
        self.token      = unpack.unpack_opaque()
        if self.major not in (const.GSS_S_COMPLETE, const.GSS_S_CONTINUE_NEEDED):
            # Display major and minor codes on error
            self.set_strfmt(2, "major: {1}, minor: {2}")
        else:
            # Try to decode the token
            krb5 = GSS_API(self.token)
            if krb5:
                # Replace token attribute with the decoded object
                self.token = krb5
                self.set_strfmt(2, "context: {0}, {4}")

class GSS_data(BaseObj):
    """struct rpc_gss_data_t {
           unsigned int    seq_num;
           proc_req_arg_t  arg;
       };
    """
    # Class attributes
    _strfmt2  = "length: {0}, seq_num: {1}"
    _attrlist = ("length", "seq_num")

    def __init__(self, unpack):
        self.length  = unpack.unpack_uint()
        self.seq_num = unpack.unpack_uint()

class GSS_checksum(GSS_init_arg): pass

class GSS(BaseObj):
    """GSS Data object

       This is a base object and should not be instantiated.
       It gives the following methods:
           # Decode data preceding the RPC payload when flavor is RPCSEC_GSS
           x.decode_gss_data()

           # Decode data following the RPC payload when flavor is RPCSEC_GSS
           x.decode_gss_checksum()
    """
    def _gss_data_call(self):
        """Internal method to decode GSS data on a CALL"""
        if self.credential.flavor != rpc_const.RPCSEC_GSS:
            # Not a GSS encoded packet
            return
        unpack = self._pktt.unpack
        if self.credential.gss_proc == const.RPCSEC_GSS_DATA:
            if self.credential.gss_service == const.rpc_gss_svc_integrity:
                return GSS_data(unpack)
        elif self.credential.gss_proc == const.RPCSEC_GSS_INIT:
            return GSS_init_arg(unpack)

    def _gss_data_reply(self):
        """Internal method to decode GSS data on a REPLY"""
        if self.verifier.flavor != rpc_const.RPCSEC_GSS and not hasattr(self.verifier, 'gss_proc'):
            # Not a GSS encoded packet
            return
        unpack = self._pktt.unpack
        if self.verifier.gss_proc == const.RPCSEC_GSS_DATA:
            if self.verifier.gss_service == const.rpc_gss_svc_integrity:
                return GSS_data(unpack)
        elif self.verifier.gss_proc == const.RPCSEC_GSS_INIT:
            return GSS_init_res(unpack)

    def decode_gss_data(self):
        """Decode GSS data"""
        try:
            pktt = self._pktt
            if pktt.unpack.size() < 4:
                # Not a GSS encoded packet
                return
            if self.type == rpc_const.CALL:
                gss = self._gss_data_call()
            else:
                gss = self._gss_data_reply()
            if gss is not None:
                pktt.pkt.gssd = gss
        except:
            pass

    def decode_gss_checksum(self):
        """Decode GSS checksum"""
        try:
            pktt = self._pktt
            unpack = pktt.unpack
            if unpack.size() < 4:
                # Not a GSS encoded packet
                return
            gss = None
            if self.type == rpc_const.CALL:
                cred = self.credential
            else:
                cred = self.verifier
            if cred.flavor == rpc_const.RPCSEC_GSS and cred.gss_proc == const.RPCSEC_GSS_DATA:
                if cred.gss_service == const.rpc_gss_svc_integrity:
                    gss = GSS_checksum(unpack)
            if gss is not None:
                pktt.pkt.gssc = gss
        except:
            pass



#===============================================================================
# Copyright 2016 NetApp, Inc. All Rights Reserved,
# contribution by Jorge Mora <mora@netapp.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#===============================================================================
"""
ARP constants module

RFC 826 An Ethernet Address Resolution Protocol
"""
import nfstest_config as c

# Module constants
__author__    = "Jorge Mora (%s)" % c.NFSTEST_AUTHOR_EMAIL
__copyright__ = "Copyright (C) 2016 NetApp, Inc."
__license__   = "GPL v2"
__version__   = "1.0"

# Enum arp_oper
REQUEST = 1
REPLY   = 2
RARP_REQUEST = 3
RARP_REPLY   = 4

arp_oper = {
    1: "REQUEST",
    2: "REPLY",
    3: "REQUEST",
    4: "REPLY",
}

# Hardware types
HTYPE_ETHERNET = 1

# Protocol types
PTYPE_IPV4 = 0x0800
PTYPE_IPV6 = 0x86dd



#===============================================================================
# Copyright 2014 NetApp, Inc. All Rights Reserved,
# contribution by Jorge Mora <mora@netapp.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#===============================================================================
# Generated by process_xdr.py from portmap2.x on Thu Sep 08 09:14:34 2016
"""
PORTMAPv2 constants module
"""
import nfstest_config as c

# Module constants
__author__    = "Jorge Mora (%s)" % c.NFSTEST_AUTHOR_EMAIL
__copyright__ = "Copyright (C) 2014 NetApp, Inc."
__license__   = "GPL v2"
__version__   = "2.0"

# Enum proto2
TCP = 6   # protocol number for TCP/IP
UDP = 17  # protocol number for UDP/IP

proto2 = {
     6 : "TCP",
    17 : "UDP",
}

# Enum portmap_proc2
PMAPPROC_NULL    = 0
PMAPPROC_SET     = 1
PMAPPROC_UNSET   = 2
PMAPPROC_GETPORT = 3
PMAPPROC_DUMP    = 4
PMAPPROC_CALLIT  = 5

portmap_proc2 = {
    0 : "PMAPPROC_NULL",
    1 : "PMAPPROC_SET",
    2 : "PMAPPROC_UNSET",
    3 : "PMAPPROC_GETPORT",
    4 : "PMAPPROC_DUMP",
    5 : "PMAPPROC_CALLIT",
}

# Enum portmap_prog2
PORTMAP   = 100000
RSTAT     = 100001
RUSERS    = 100002
NFS       = 100003
YPSERV    = 100004
MOUNT     = 100005
RDBX      = 100006
YPBIND    = 100007
WALL      = 100008
YPPASSWDD = 100009
ETHERSTAT = 100010
RQUOTA    = 100011
REXEC     = 100017
NLOCKMGR  = 100021
STATMON1  = 100023
STATMON2  = 100024
YPUPDATE  = 100028
NFS_ACL   = 100227

portmap_prog2 = {
    100000 : "PORTMAP",
    100001 : "RSTAT",
    100002 : "RUSERS",
    100003 : "NFS",
    100004 : "YPSERV",
    100005 : "MOUNT",
    100006 : "RDBX",
    100007 : "YPBIND",
    100008 : "WALL",
    100009 : "YPPASSWDD",
    100010 : "ETHERSTAT",
    100011 : "RQUOTA",
    100017 : "REXEC",
    100021 : "NLOCKMGR",
    100023 : "STATMON1",
    100024 : "STATMON2",
    100028 : "YPUPDATE",
    100227 : "NFS_ACL",
}



#===============================================================================
# Copyright 2012 NetApp, Inc. All Rights Reserved,
# contribution by Jorge Mora <mora@netapp.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#===============================================================================
import os

NFSTEST_PACKAGE          = 'NFStest'
NFSTEST_VERSION          = '2.1.5'
NFSTEST_SUMMARY          = 'NFS Test Suite'
NFSTEST_AUTHOR           = 'Jorge Mora'
NFSTEST_AUTHOR_EMAIL     = 'mora@netapp.com'
NFSTEST_MAINTAINER       = NFSTEST_AUTHOR
NFSTEST_MAINTAINER_EMAIL = NFSTEST_AUTHOR_EMAIL
NFSTEST_COPYRIGHT        = "Copyright (C) 2012 NetApp, Inc."
NFSTEST_LICENSE          = 'GPLv2'
NFSTEST_URL    = 'http://wiki.linux-nfs.org/wiki/index.php/NFStest'
NFSTEST_DL_URL = 'http://www.linux-nfs.org/~mora/nfstest/releases/nfstest.tgz'
NFSTEST_DESCRIPTION = '''NFS Test Suite

Provides a set of tools for testing either the NFS client or the NFS server,
included tests focused mainly on testing the client. These tools include the
following:

Test utilities package (nfstest)
===============================
Provides a set of tools for testing either the NFS client or the NFS server,
most of the functionality is focused mainly on testing the client.
These tools include the following:

    - Process command line arguments
    - Provide functionality for PASS/FAIL
    - Provide test grouping functionality
    - Provide multiple client support
    - Logging mechanism
    - Debug info control
    - Mount/Unmount control
    - Create files/directories
    - Provide mechanism to start a packet trace
    - Provide mechanism to simulate a network partition
    - Support for pNFS testing

Packet trace package (packet)
============================
The Packet trace module takes a trace file created by tcpdump and unpacks
the contents of each packet. You can decode one packet at a time, or do a
search for specific packets. The main difference between this modules and
other tools used to decode trace files is that you can use this module to
completely automate your tests.

Packet layers supported:
    - Ethernet II (RFC 894)
    - IP layer (supports v4 only)
    - TCP layer
    - RPC layer
    - NFS v4.0
    - NFS v4.1 including pNFS file layouts
'''

NFSTEST_MAN_MAP = {}
def _get_manpages(src_list, mandir, section, mod=False):
    manpages = []
    for src in src_list:
        if src == 'README':
            manpage = os.path.join(mandir, 'nfstest.%d.gz' % section)
        elif mod:
            if '__init__' in src:
                continue
            manpage = os.path.splitext(src.replace('/', '.'))[0]
            manpage = os.path.join(mandir, manpage + '.%d.gz' % section)
        else:
            manpage = os.path.split(src)[1]
            manpage = os.path.join(mandir, manpage + '.%d.gz' % section)
        manpages.append(manpage)
        NFSTEST_MAN_MAP[src] = manpage
    return manpages

bin_dirs = [
    '/usr/bin',
    '/usr/sbin',
    '/bin',
    '/sbin',
]
def _find_exec(command):
    for bindir in bin_dirs:
        bincmd = os.path.join(bindir, command)
        if os.path.exists(bincmd):
            return bincmd
    return command

NFSTEST_TESTDIR = 'test'
NFSTEST_MANDIR  = 'man'
NFSTEST_USRMAN  = '/usr/share/man'
NFSTEST_CONFIG  = '/etc/nfstest'
NFSTEST_HOMECFG = os.path.join(os.environ.get('HOME',''), '.nfstest')
NFSTEST_CWDCFG  = '.nfstest'
NFSTEST_SCRIPTS = [
    'test/nfstest_alloc',
    'test/nfstest_cache',
    'test/nfstest_delegation',
    'test/nfstest_dio',
    'test/nfstest_file',
    'test/nfstest_interop',
    'test/nfstest_io',
    'test/nfstest_lock',
    'test/nfstest_pkt',
    'test/nfstest_pnfs',
    'test/nfstest_posix',
    'test/nfstest_sparse',
    'test/nfstest_ssc',
    'test/nfstest_xid',
]
NFSTEST_ALLMODS = [
    'baseobj.py',
    'formatstr.py',
    'nfstest/file_io.py',
    'nfstest/host.py',
    'nfstest/nfs_util.py',
    'nfstest/rexec.py',
    'nfstest/test_util.py',
    'nfstest/utils.py',
    'packet/derunpack.py',
    'packet/pkt.py',
    'packet/pktt.py',
    'packet/record.py',
    'packet/unpack.py',
    'packet/utils.py',
    'packet/application/dns.py',
    'packet/application/dns_const.py',
    'packet/application/gss.py',
    'packet/application/gss_const.py',
    'packet/application/krb5.py',
    'packet/application/krb5_const.py',
    'packet/application/ntp4.py',
    'packet/application/rpc.py',
    'packet/application/rpc_const.py',
    'packet/application/rpc_creds.py',
    'packet/internet/arp.py',
    'packet/internet/arp_const.py',
    'packet/internet/ipv4.py',
    'packet/internet/ipv6.py',
    'packet/internet/ipv6addr.py',
    'packet/link/ethernet.py',
    'packet/link/macaddr.py',
    'packet/nfs/mount3.py',
    'packet/nfs/mount3_const.py',
    'packet/nfs/nfs3.py',
    'packet/nfs/nfs3_const.py',
    'packet/nfs/nfs4.py',
    'packet/nfs/nfs4_const.py',
    'packet/nfs/nfs.py',
    'packet/nfs/nfsbase.py',
    'packet/nfs/nlm4.py',
    'packet/nfs/nlm4_const.py',
    'packet/nfs/portmap2.py',
    'packet/nfs/portmap2_const.py',
    'packet/transport/tcp.py',
    'packet/transport/udp.py',
]

NFSTEST_MAN1  = _get_manpages(['README'], NFSTEST_MANDIR, 1)
NFSTEST_MAN1 += _get_manpages(NFSTEST_SCRIPTS, NFSTEST_MANDIR, 1)
NFSTEST_MAN3  = _get_manpages(NFSTEST_ALLMODS, NFSTEST_MANDIR, 3, mod=True)
NFSTEST_MODULES = ['baseobj', 'formatstr', 'nfstest_config']
NFSTEST_PACKAGES = [
    'nfstest',
    'packet',
    'packet.application',
    'packet.internet',
    'packet.link',
    'packet.nfs',
    'packet.transport',
]

# Default values
NFSTEST_NFSVERSION   = 4.1
NFSTEST_NFSPROTO     = 'tcp'
NFSTEST_NFSPORT      = 2049
NFSTEST_NFSSEC       = 'sys'
NFSTEST_EXPORT       = '/'
NFSTEST_MTPOINT      = '/mnt/t'
NFSTEST_MTOPTS       = 'hard,rsize=4096,wsize=4096'
NFSTEST_INTERFACE    = 'eth0'
NFSTEST_SUDO         = _find_exec('sudo')
NFSTEST_IPTABLES     = _find_exec('iptables')
NFSTEST_TCPDUMP      = _find_exec('tcpdump')
NFSTEST_MESSAGESLOG  = '/var/log/messages'
NFSTEST_TMPDIR       = '/tmp'



#!/usr/bin/env python
#===============================================================================
# Copyright 2012 NetApp, Inc. All Rights Reserved,
# contribution by Jorge Mora <mora@netapp.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#===============================================================================
import os
import re
import sys
import time
import tokenize
import StringIO
import subprocess
import nfstest_config as c

# Module constants
__author__    = "Jorge Mora (%s)" % c.NFSTEST_AUTHOR_EMAIL
__copyright__ = "Copyright (C) 2012 NetApp, Inc."
__license__   = "GPL v2"
__version__   = "1.2"

def _get_modules(script):
    # Read the whole file
    with open(script, "r") as fd:
        filedata = fd.read()

    # Join code lines separated by "\" at the end of the line
    # because untokenize fails with split code lines
    filedata = re.sub(r"\\\n\s+", r" ", filedata)

    # Have the file data be used as a file
    fd = StringIO.StringIO(filedata)

    # Remove all comments and replace strings so all matches are done
    # on the source code only
    tokenlist = []
    for tok in tokenize.generate_tokens(fd.readline):
        toktype, tok_string, start, end, line = tok
        if toktype == tokenize.COMMENT:
            # Remove all comments
            tok = (toktype, "", start, end, line)
        elif toktype == tokenize.STRING:
            # Replace all strings
            tok = (toktype, "'STRING'", start, end, line)
        tokenlist.append(tok)
    filedata = tokenize.untokenize(tokenlist)
    fd.close()

    modules = {}
    for line in filedata.split("\n"):
        line = line.lstrip().rstrip()
        m = re.search(r'^(from|import)\s+(.*)', line)
        if m:
            mods = m.group(2)
            mods = mods.split(' as ')[0]
            modlist = mods.split(' import ')
            mod_entries = []
            for mods in modlist:
                mods = mods.split(',')
                mod_entries.append([])
                for item in mods:
                    mod_entries[-1].append(item.strip())
            if mod_entries:
                for mods in mod_entries[0]:
                    modules[mods] = 1
                if len(mod_entries) > 1:
                    for mods in mod_entries[0]:
                        for item in mod_entries[1]:
                            modules['.'.join([mods, item])] = 1
    return modules.keys()

def _get_see_also(src, manpage, modules, local_mods):
    parent_objs = {}
    dirname = os.path.dirname(os.path.abspath(src))
    for item in modules:
        if item not in local_mods and item[0] != '_':
            if item.find(".") < 0:
                # This module has only one component, check if it is on the
                # same directory as the source
                itempath = os.path.join(dirname, item+".py")
                if os.path.exists(itempath):
                    items = manpage.split(".")
                    if len(items) > 2:
                        item = ".".join(items[:-2] + [item])
            osrc = item.replace('.', '/')
            osrcpy = osrc + '.py'
            if src in (osrc, osrcpy):
                continue
            mangz = c.NFSTEST_MAN_MAP.get(osrc) or c.NFSTEST_MAN_MAP.get(osrcpy)
            obj = ".BR %s" % os.path.split(item)[1]
            if mangz:
                m = re.search(r'([^\.]+)\.gz$', mangz)
                if m:
                    obj += "(%s)" % m.group(1)
                    parent_objs[obj] = 1
    objs = parent_objs.keys()
    objs.sort()
    return ',\n'.join(objs)

def _check_script(script):
    fd = open(script, 'r')
    line = fd.readline()
    fd.close()
    if re.search('^#!.*python', line):
        return True
    return False

def _lstrip(lines, br=False):
    ret = []
    minsps = 99999
    for line in lines:
        # Ignore blank lines
        if len(line) == 0:
            continue
        nsp = len(line) - len(line.lstrip())
        minsps = min(minsps, nsp)
    for line in lines:
        line = line[minsps:]
        if len(line.lstrip()) > 0:
            if br and line.lstrip()[0] in ('#', '$', '%'):
                ret.append('.br')
            if line[0] in ("'", '"'):
                line = '\\t' + line
        ret.append(line)
    while len(ret) and ret[-1] == "":
        ret.pop()
    return ret

def _process_func(lines):
    ret = []
    in_arg = False
    need_re = False
    count = 0
    for line in _lstrip(lines):
        if re.search(r'^[a-z]\w*:', line):
            if not in_arg:
                # Start indented region
                ret.append('.RS')
                need_re = True
            ret.append('.TP\n.B')
            in_arg = True
        elif len(line) == 0:
            if in_arg:
                # End of indented region
                ret.append('.RE\n.RS')
            in_arg = False
        elif in_arg:
            line = line.lstrip()
        if len(line) and line[0] == '#':
            count += 1
        ret.append(line)
    if count >= len(ret) - 1:
        ret_new = []
        for line in ret:
            ret_new.append(line.lstrip('#'))
        ret = ret_new
    if need_re:
        ret.append('.RE')
    return ret

def create_manpage(src, dst):
    usage = ''
    summary = ''
    desc_lines = []
    description = ''
    author = '%s (%s)' % (c.NFSTEST_AUTHOR, c.NFSTEST_AUTHOR_EMAIL)
    notes = []
    examples = []
    bugs = ''
    see_also = ''
    version = ''
    classes = []
    func_list = []
    test = {}
    tests = []
    tool = {}
    tools = []
    option = {}
    options = []
    section = ''
    dlineno = 0
    requirements = []
    installation = []
    progname = ''

    is_script = _check_script(src)

    if not os.path.isdir(dst):
        manpage = dst
    elif is_script:
        manpage = os.path.join(dst, os.path.splitext(os.path.split(src)[1])[0] + '.1')
    else:
        manpage = os.path.splitext(src)[0].replace('/', '.') + '.3'
        manpage = manpage.lstrip('.')
    manpagegz = manpage + '.gz'

    fst = os.stat(src)
    if os.path.exists(manpagegz) and fst.st_mtime < os.stat(manpagegz).st_mtime:
        return

    print "Creating man page for %s" % src
    modules = _get_modules(src)

    if src == 'README':
        fd = open(src, 'r')
        lines = []
        for line in fd.readlines():
            lines.append(line.rstrip())
        fd.close()
        progname = 'NFStest'
    elif is_script:
        cmd = "%s --version" % src
        proc = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        pstdout, pstderr = proc.communicate()
        proc.wait()
        version = pstdout.split()[1]

        cmd = "%s --help" % src
        proc = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        pstdout, pstderr = proc.communicate()
        proc.wait()
        lines = re.sub('Total time:.*', '', pstdout)
        lines = re.sub('TIME:\s+[0-9.]+s.*', '', lines)
        lines = re.sub('0 tests \(0 passed, 0 failed\)', '', lines)
        lines = lines.split('\n')
        while lines[-1] == "":
            lines.pop()
    else:
        absmodule = os.path.splitext(src)[0].replace('/', '.')
        cmd = "pydoc %s" % absmodule
        proc = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        pstdout, pstderr = proc.communicate()
        proc.wait()
        lines = pstdout.split('\n')

    for line in lines:
        if is_script and len(usage) == 0:
            m = re.search(r'^Usage:\s+(.*)', line)
            usage = m.group(1)
            continue
        elif len(summary) == 0:
            if len(line) > 0:
                if re.search(r'^FILE', line):
                    summary = ' '
                    continue
                else:
                    summary = ' - ' + line
                section = 'description'
            continue
        elif len(line) > 0 and line[0] == '=':
            continue
        elif re.search(r'^Requirements and limitations', line):
            section = 'requirements'
            continue
        elif re.search(r'^Tests', line):
            section = 'tests'
            continue
        elif re.search(r'^Tools', line):
            section = 'tools'
            continue
        elif re.search(r'^Installation', line):
            section = 'installation'
            continue
        elif re.search(r'^Run the tests', line):
            section = 'examples'
            continue
        elif re.search(r'^Useful options', line):
            section = 'options'
            continue
        elif re.search(r'^Examples:', line):
            section = 'examples'
            continue
        elif re.search(r'^Notes:', line):
            section = 'notes'
            continue
        elif re.search(r'^Available tests:', line):
            section = 'tests'
            continue
        elif re.search(r'^Options:', line):
            section = 'options'
            continue
        elif re.search(r'^NAME', line):
            section = 'name'
            continue
        elif re.search(r'^DESCRIPTION', line):
            section = 'desc'
            continue
        elif re.search(r'^CLASSES', line):
            section = 'class'
            continue
        elif re.search(r'^FUNCTIONS', line):
            section = 'funcs'
            continue
        elif re.search(r'^DATA', line):
            section = 'data'
            continue
        elif re.search(r'^VERSION', line):
            section = 'version'
            continue
        elif re.search(r'^AUTHOR', line):
            section = 'author'
            continue

        if section == 'name':
            section = ''
            m = re.search(r'^\s*(\S+)(.*)', line)
            progname = m.group(1)
            summary = m.group(2)
        elif section == 'desc':
            desc_lines.append(line)
        elif section == 'description':
            if progname == 'NFStest':
                if re.search(r'^\s*=+', line):
                    if dlineno == 0:
                        dlineno = len(desc_lines) - 1
                    desc_lines[-1] = re.sub(r'^(\s*)', r'\1.SS ', desc_lines[-1])
                else:
                    desc_lines.append(line)
            else:
                description += line + '\n'
        elif section == 'requirements':
            requirements.append(line)
        elif section == 'examples':
            examples.append(line)
        elif section == 'notes':
            notes.append(line)
        elif section == 'tests':
            if progname == 'NFStest':
                if re.search(r'^\s*=+', line):
                    continue
                testname = re.search(r'\s*(\w+)\s+-', line)
            else:
                testname = re.search(r'\s*(.*):$', line)
            if testname:
                if test:
                    tests.append(test)
                    test = {}
                test['name'] = testname.group(1)
                test['desc'] = []
            else:
                test['desc'].append(line)
        elif section == 'tools':
            if progname == 'NFStest':
                if re.search(r'^\s*=+', line):
                    continue
                toolname = re.search(r'\s*(\w+)\s+-', line)
            else:
                toolname = re.search(r'\s*(.*):$', line)
            if toolname:
                if tool:
                    tools.append(tool)
                    tool = {}
                tool['name'] = toolname.group(1)
                tool['desc'] = []
            else:
                tool['desc'].append(line)
        elif section == 'installation':
            installation.append(line)
        elif section == 'options':
            if progname == 'NFStest':
                optsname = re.search(r'^(((-\w(\s+\S+)?),\s+)?--.+)', line)
            else:
                optsname = re.search(r'^\s*(((-\w(\s+\S+)?),\s+)?--(\S+))\s*(.*)', line)
            if optsname:
                if option:
                    options.append(option)
                    option = {}
                option['name'] = optsname.group(1)
                if len(optsname.groups()) >= 6 and len(optsname.group(6)) > 0:
                    option['desc'] = [optsname.group(6)]
                else:
                    option['desc'] = []
            else:
                if progname == 'NFStest':
                    option['desc'].append(line)
                else:
                    if line[0:4] == "    ":
                        option['desc'].append(line.lstrip())
                    else:
                        option['group'] = line.lstrip()
        elif section == 'class':
            line = line.lstrip().lstrip('|')
            classes.append(line)
        elif section == 'funcs':
            func_list.append(line)
        elif section == 'version':
            section = ''
            version = line.lstrip()
        elif section == 'author':
            section = ''
            author = line.lstrip()

    if test and section != 'tests':
        tests.append(test)
        test = {}
    if tool and section != 'tests':
        tools.append(tool)
        tool = {}

    class_list = []
    if classes:
        # Process all classes
        for line in classes:
            # Class definition:
            #     class classname(prototype)
            # or a copy of different class:
            #     classname = class sourceclass(prototype)
            m = re.search(r'^((\w+)\s+=\s+)?class\s+(\w+)(.*)', line)
            if m:
                data = m.groups()
                if data[1] is None:
                    copy = None
                    cls_name = data[2]
                else:
                    copy = data[2]
                    cls_name = data[1]
                class_list.append({'name': cls_name, 'proto': data[3], 'body': [], 'res': [], 'copy': copy})
            elif class_list:
                class_list[-1]['body'].append(line)
        for cls in class_list:
            body = []
            method_desc = []
            in_methods = False
            in_inherit = False
            in_resolution = False
            for line in _lstrip(cls['body']):
                if re.search(r'^Data descriptors defined here:', line):
                    break
                if len(line) > 1 and line == '-' * len(line):
                    continue
                elif re.search(r'^Method resolution order:', line):
                    in_resolution = True
                    in_methods = False
                elif re.search(r'^(Static )?[mM]ethods inherited', line):
                    in_inherit = True
                    in_methods = False
                elif re.search(r'^(Static )?[mM]ethods defined here:', line):
                    body += _process_func(method_desc)
                    method_desc = []
                    body.append('.P\n.B %s\n%s' % (line, '-' * len(line)))
                    in_methods = True
                elif in_methods and re.search(r'^\w+(\s+=\s+\w+)?\(', line):
                    body += _process_func(method_desc)
                    method_desc = []
                    body.append('.P\n.B %s' % line)
                elif in_methods:
                    method_desc.append(line)
                elif in_resolution:
                    if len(line) == 0:
                        in_resolution = False
                    else:
                        cls['res'].append(line.lstrip())
                elif not in_inherit and not in_resolution:
                    body.append(line)
            body += _process_func(method_desc)
            cls['body'] = body

    all_modules = modules
    local_mods = []
    for cls in class_list:
        if cls['body']:
            mods = []
            for item in cls['res']:
                mods.append(item)
                obj = '.'.join(item.split('.')[:-1])
                if len(obj):
                    mods.append(obj)
            all_modules += mods
            local_mods.append(cls['name'])
    all_modules += c.NFSTEST_SCRIPTS if is_script or progname == 'NFStest' else []
    see_also += _get_see_also(src, manpage, all_modules, local_mods)

    # Get a list of functions included from imported modules
    mod_funcs = []
    for mod in modules:
        data = mod.split(".")
        if len(data) > 1:
            mod_funcs.append(data[-1])

    func_desc = []
    functions = []
    is_local_function = False
    for line in _lstrip(func_list):
        regex = re.search(r'(\w+)\((.*)\)', line)
        if not regex:
            regex = re.search(r'(\w+)\s+(lambda)\s+(.*)', line)
        if regex:
            data = regex.groups()
            if len(data) == 3:
                line = "%s(%s)" % (data[0], data[2])
            is_local_function = False
            functions += _process_func(func_desc)
            func_desc = []
            if data[1] != "..." or data[0] not in mod_funcs:
                # Only include functions defined locally,
                # do not include any function from imported modules
                functions.append('.SS %s' % line)
                is_local_function = True
        elif is_local_function:
            func_desc.append(line)
    functions += _process_func(func_desc)

    if option:
        options.append(option)

    if progname == 'NFStest':
        description += '\n'.join(_lstrip(desc_lines[:dlineno]))
        description += '\n'.join(_lstrip(desc_lines[dlineno:]))
    elif desc_lines:
        description += '\n'.join(_lstrip(desc_lines))

    if is_script:
        progname = os.path.splitext(usage.split()[0])[0]

    pname = progname.split('.')[-1]
    datestr = time.strftime("%e %B %Y")

    # Open man page to create
    fd = open(manpage, 'w')

    thisprog = os.path.split(sys.argv[0])[1]
    print >>fd, '.\\" DO NOT MODIFY THIS FILE!  It was generated by %s %s.' % (thisprog, __version__)
    nversion = "%s %s" % (c.NFSTEST_PACKAGE, c.NFSTEST_VERSION)
    if is_script or progname == 'NFStest':
        man_section = 1
    else:
        man_section = 3
    print >>fd, '.TH %s %d "%s" "%s" "%s %s"' % (pname.upper(), man_section, datestr, nversion, pname, version)
    print >>fd, '.SH NAME'
    print >>fd, '%s%s' % (progname, summary)
    if len(usage):
        print >>fd, '.SH SYNOPSIS'
        print >>fd, usage
    if len(description) and description != '\n':
        print >>fd, '.SH DESCRIPTION'
        print >>fd, description
    if requirements:
        print >>fd, '.SH REQUIREMENTS AND LIMITATIONS'
        print >>fd, '\n'.join(_lstrip(requirements))
    if class_list:
        print >>fd, '.SH CLASSES'
        for cls in class_list:
            if cls['body'] and cls['copy']:
                print >>fd, ".SS class %s%s" % (cls['name'], cls['proto'])
                print >>fd, ".nf\n%s = class %s%s\n.fi" % (cls['name'], cls['copy'], cls['proto'])
            elif cls['body']:
                print >>fd, ".SS class %s%s\n.nf" % (cls['name'], cls['proto'])
                for line in cls['body']:
                    print >>fd, line
                print >>fd, ".fi"
    if functions:
        print >>fd, '.SH FUNCTIONS'
        for line in functions:
            print >>fd, line
    if options and progname != 'NFStest':
        print >>fd, '.SH OPTIONS'
        for option in options:
            print >>fd, '.IP "%s"' % option['name']
            print >>fd, '\n'.join(_lstrip(option['desc']))
            if option.get('group'):
                print >>fd, '\n.SS %s\n' % option['group']

    if tests:
        print >>fd, '.SH TESTS'
        for test in tests:
            print >>fd, '.SS %s\n.nf' % test['name']
            print >>fd, '\n'.join(_lstrip(test['desc']))
            print >>fd, '.fi'

    if tools:
        print >>fd, '.SH TOOLS'
        for tool in tools:
            print >>fd, '.SS %s\n.nf' % tool['name']
            print >>fd, '\n'.join(_lstrip(tool['desc']))
            print >>fd, '.fi'

    if installation:
        print >>fd, '.SH INSTALLATION'
        print >>fd, '\n'.join(_lstrip(installation))

    if examples:
        print >>fd, '.SH EXAMPLES'
        print >>fd, '\n'.join(_lstrip(examples, br=True))

    if options and progname == 'NFStest':
        print >>fd, '.SH USEFUL OPTIONS'
        for option in options:
            print >>fd, '.IP "%s"' % option['name']
            print >>fd, '\n'.join(_lstrip(option['desc']))
    if notes:
        print >>fd, '.SH NOTES'
        print >>fd, '\n'.join(_lstrip(notes))

    if len(see_also) > 0:
        print >>fd, '.SH SEE ALSO'
        print >>fd, see_also + "\n"

    print >>fd, '.SH BUGS'
    if len(bugs) > 0:
        print >>fd, bugs
    else:
        print >>fd, 'No known bugs.'

    print >>fd, '.SH AUTHOR'
    print >>fd, author
    fd.close()
    cmd = "gzip -f --stdout %s > %s.gz" % (manpage, manpage)
    os.system(cmd)

def run():
    if not os.path.exists(c.NFSTEST_MANDIR):
        os.mkdir(c.NFSTEST_MANDIR)
    for (script, manpagegz) in c.NFSTEST_MAN_MAP.items():
        manpage = os.path.splitext(manpagegz)[0]
        create_manpage(script, manpage)

######################################################################
# Entry
if __name__ == '__main__':
    if len(sys.argv) > 1:
        dir = sys.argv[2] if len(sys.argv) == 3 else '.'
        create_manpage(sys.argv[1], dir)
    else:
        run()
